package org.processmining.resourcerecommendation.algorithms;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import lpsolve.LpSolve;
import lpsolve.LpSolveException;

import org.processmining.resourcerecommendation.models.Dimension;
import org.processmining.resourcerecommendation.models.Dimensions;
import org.processmining.resourcerecommendation.models.RecommendationElement;
import org.processmining.resourcerecommendation.models.Request;
import org.processmining.resourcerecommendation.models.Resource;
import org.processmining.resourcerecommendation.models.ResourceKnowledge;
import org.processmining.resourcerecommendation.models.ResourceRecommendation;
import org.processmining.resourcerecommendation.models.WorkloadDimension;
import org.processmining.resourcerecommendation.models.impl.ResourceProcessFactory;
import org.processmining.resourcerecommendation.parameters.RecommendResourcesParameters;

public class AllocationAlgorithm {

	public AllocationAlgorithm() {
		System.loadLibrary("lpsolve55j");
	}

	public ResourceRecommendation apply(ResourceKnowledge knowledge, RecommendResourcesParameters parameters) throws LpSolveException {
		if (parameters.getRequests().size() == 1){
			//Recommendation On-demand with LP(one requests)
			return recommendAssignment(knowledge, parameters);
		}
		else if (parameters.getRequests().size() > 1){
				//Recommendation BATCH with LP(several requests at the same time)
				return recommendAssignment(knowledge, parameters);
				}
		else{
			//No Recommendation: empty recommendation
			ResourceRecommendation recommendation = ResourceProcessFactory.createResourceRecommendation();
			return recommendation;
		}
	}

	public ResourceRecommendation recommendAssignment(ResourceKnowledge knowledge, RecommendResourcesParameters parameters) throws LpSolveException {
		LpSolve lp;
		List<Resource> resources = knowledge.getResources();
		List<Dimension> dimensions = knowledge.getDimensions();
		Map<String, Double> weights = parameters.getWeights();

		int nCol, ret = 0;  
		int numRequests = parameters.getRequests().size();
		int numResources = resources.size();
		int[] maxResourceUsage = new int[numResources];
		double[][] values = new double[numRequests][numResources];
		double sum = 0;

		ResourceRecommendation recom = ResourceProcessFactory.createResourceRecommendation();

		for (int j=0; j<numResources; j++) {
			String resourceId = resources.get(j).getId();
			for(int d=0; d<dimensions.size(); d++){
				String dimensionId = dimensions.get(d).getId();
				if (dimensionId == Dimensions.WORKLOAD) {
					WorkloadDimension wd = (WorkloadDimension) dimensions.get(d);
					maxResourceUsage[j] = wd.getTop(resourceId) - wd.getCurrent(resourceId);					
				}
			}
		}
		
		for(int j=0; j<numRequests; j++){
			Request request = parameters.getRequests().get(j); //Multiple requests
			for(int pos=0; pos < resources.size(); pos++){
				sum=0;
				for(int d=0; d<dimensions.size(); d++){
					String characterizationId = request.getCharacterizationId();
					String resourceId = resources.get(pos).getId();
					String dimensionId = dimensions.get(d).getId();

					//if (dimensionId != Dimensions.WORKLOAD) { // Consider workload to compute the overall score
						double metric = knowledge.getMetric(dimensionId, characterizationId, resourceId);
						double weight = weights.get(dimensionId);
						double score = metric * weight;	
						sum = sum + score;
					//}
				}
				double avg = sum/dimensions.size();
				values[j][pos]= avg;
				System.out.println(values[j][pos]);
			}
		}
		
		/* We will build the model row by row
	         So we start with creating a model with 0 rows and Ncol columns */
		nCol = numRequests*numResources; /* there are Ncol variables in the model */

		/* create space large enough for one row */
		int[] colno = new int[nCol];
		double[] row;
		double[] zeroRow = new double[nCol];
		for (int c=0; c<nCol; c++){
			colno[c] = c+1;
			zeroRow[c] = 0;
		}
		lp = LpSolve.makeLp(0, nCol);
		if(lp.getLp() == 0)
			ret = 1; /* couldn't construct a new model... */
		if(ret == 0) {
			/* let us name our variables. Not required, but can be useful for debugging */
			for (int i=0; i<numRequests; i++){
				for (int j=0; j<numResources; j++){
					String s = "x_" + (i+1) + "_" + (j+1);
					lp.setColName(i*numResources+j+1, s);
					lp.setInt(i*numResources+j+1, true);
				}
			}
			lp.setAddRowmode(true);  /* makes building the model faster if it is done row by row */

			//int nRow = 0;

			/* construct first nRequests rows (sum Xij = 1, j=1..nResources; for all i = 1..nRequests)*/
			for (int i=0; i<numRequests; i++){
				row = zeroRow.clone();
				for (int j=0; j<numResources; j++)
					row[i*numResources+j] = 1;
				/* add the row to lpsolve*/
				lp.addConstraintex(nCol, row, colno, LpSolve.EQ, 1);
			}
			/* construct next nResources rows (sum Xij = 1, j=1..nResources; for all i = 1..nRequests)*/
			for (int j=0; j<numResources; j++){
				row = zeroRow.clone();
				for (int i=0; i<numRequests; i++)
					row[i*numResources+j] = 1;
				/* add the row to lpsolve */
				lp.addConstraintex(nCol, row, colno, LpSolve.LE, maxResourceUsage[j]);
			}

			lp.setAddRowmode(false); /* rowmode should be turned off again when done building the model */

			/* set the objective function (X ij * Value ij) */
			row = zeroRow.clone();
			for (int i=0; i<numRequests; i++){
				for (int j=0; j<numResources; j++)
					row[i*numResources+j] = values[i][j];
			}
			/* set the objective in lpsolve */
			lp.setObjFnex(nCol, row, colno);

			/* set the object direction to maximize */
			lp.setMaxim();

			/* just out of curiosity, now generate the model in lp format in file model.lp */
			lp.writeLp("model.lp");

			/* I only want to see important messages on screen while solving */
			lp.setVerbose(LpSolve.IMPORTANT);

			/* Now let lpsolve calculate a solution */
			ret = lp.solve();
			if(ret == LpSolve.OPTIMAL)
				ret = 0;
			else
				ret = 5;
		}

		if(ret == 0) {
			/* a solution is calculated, now lets get some results */

			/* objective value */
			System.out.println("Objective value: " + lp.getObjective());

			/* variable values */
			row = zeroRow.clone();
			lp.getVariables(row);
			for(int c = 0; c < nCol; c++)
				System.out.println(lp.getColName(c + 1) + ": " + row[c]);

			int[] solution = new int[numRequests];
			for (int i=0; i<numRequests; i++){
				for (int j=0; j<numResources; j++){
					if ((int)row[i*numResources+j] == 1){
						solution[i] = j;
					}
				}
			}
			for(int i=0; i<numRequests; i++){
				Request request = parameters.getRequests().get(i); //Multiple requests
				List<RecommendationElement> ranking = new ArrayList<RecommendationElement>();

				Resource resource = resources.get(solution[i]);
				RecommendationElement elem = ResourceProcessFactory.createRecommendationElement(resource.getId(),resource.getDescription());
				double overallScore = values[i][solution[i]];
				elem.setOverallScore(overallScore);

				for(int d=0; d<dimensions.size(); d++){
					String resourceId = resource.getId();
					String dimensionId = dimensions.get(d).getId();
					if (dimensionId != Dimensions.WORKLOAD) {
						double metric = knowledge.getMetric(dimensionId, request.getCharacterizationId(), resourceId);
						double weight = weights.get(dimensions.get(d).getId());
						elem.setScore(dimensionId, metric * weight);
					}
				}
				ranking.add(elem);

				recom.setRanking(request, ranking);
			}
			/* we are done now */
		}

		/* clean up such that all used memory by lpsolve is freed */
		if(lp.getLp() != 0)
			lp.deleteLp();

		return(recom);
	}	      
}
