package org.processmining.resourcerecommendation.algorithms;

import java.util.Map;
import java.util.Map.Entry;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XCostExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XOrganizationalExtension;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.resourcerecommendation.models.Characterization;
import org.processmining.resourcerecommendation.models.Expertise;
import org.processmining.resourcerecommendation.models.Factor;
import org.processmining.resourcerecommendation.models.FactorValue;
import org.processmining.resourcerecommendation.models.Resource;
import org.processmining.resourcerecommendation.models.ResourceKnowledge;
import org.processmining.resourcerecommendation.models.Typology;
import org.processmining.resourcerecommendation.models.Unit;
import org.processmining.resourcerecommendation.models.impl.ResourceProcessFactory;
import org.processmining.resourcerecommendation.parameters.GenerateResourceKnowledgeParameters;
import org.processmining.resourcerecommendation.utils.resrecxes.XResRecExtension;

public class GenerateResourceKnowledgeAlgorithm {

	public ResourceKnowledge apply(PluginContext context, XLog log, GenerateResourceKnowledgeParameters parameters) {
		return apply(log, parameters);
	}

	public ResourceKnowledge apply(XLog log, GenerateResourceKnowledgeParameters parameters) {
		//Object to construct
		ResourceKnowledge result = ResourceProcessFactory.createResourceKnowledge();
		
		
		//Resource Information
		for (Entry<String,String> r: XResRecExtension.instance().extractResources(log).entrySet()){
			Resource resource = ResourceProcessFactory.createResource(r.getKey(), r.getValue());
			result.addResource(resource);
		}
		
		//Units Information
		for (Entry<String,String> u: XResRecExtension.instance().extractUnits(log).entrySet()){
			Unit unit = ResourceProcessFactory.createUnit(u.getKey(), u.getValue());
			result.addUnit(unit);
		}
		
		
		//Factors Information
		for(Entry<Pair<String,String>,Map<String,String>> f: XResRecExtension.instance().extractFactors(log).entrySet()){
			String factorId = f.getKey().getFirst();
			String factorDescription = f.getKey().getSecond();
			Factor factor = ResourceProcessFactory.createFactor(factorId, factorDescription);
			for(Entry<String,String> o: f.getValue().entrySet()){
				String oId = o.getKey();
				String oDescription = o.getValue();
				FactorValue factorValue = ResourceProcessFactory.createFactorValue(oId, oDescription);
				factor.addValue(factorValue);
			}
			result.addFactor(factor);
		}
		
		
		//Typologies Information
		for(Entry<Pair<String,String>,Map<String,String>> t: XResRecExtension.instance().extractTypologies(log).entrySet()){
			String tId = t.getKey().getFirst();
			String tDescription = t.getKey().getSecond();
			Typology typology = ResourceProcessFactory.createTypology(tId, tDescription);
			for(Entry<String,String> o: t.getValue().entrySet()){
				String oId = o.getKey();
				String oDescription = o.getValue();
				typology.addFactor(oId, oDescription);
			}
			result.addTypology(typology);
		}
		
		
		//Characterization Information
		for(Entry<Pair<String,String>,Pair<String,String>> entry: XResRecExtension.instance().extractCharacterizations(log).entrySet()){
			String characterizationId = entry.getKey().getFirst();
			String characterizationDescription = entry.getKey().getSecond();
			String unitId = entry.getValue().getFirst();
			String typologyId =  entry.getValue().getSecond();
			Characterization characterization = ResourceProcessFactory.createCharacterization(characterizationId, characterizationDescription, unitId, typologyId);
			result.addCharacterization(characterization);
		}
		
		//Expertise Information
		for (Entry<String,String> e: XResRecExtension.instance().extractExpertises(log).entrySet()){
			Expertise expertise = ResourceProcessFactory.createExpertise(e.getKey(), e.getValue());
			result.addExpertise(expertise);
		}
		
		//Expertise: OverQualitifation / UnderQualification
		for (Entry<Pair<String,String>, Map<String, Long>> c: XResRecExtension.instance().extractExpertiseRequired(log).entrySet()){
			String cId = c.getKey().getFirst();
			for (Entry<String, Long> e: c.getValue().entrySet()){
				String eId = e.getKey();
				long value = e.getValue();
				result.addRequiredExpertise(cId, eId, value);
			}
			
		}
		for (Entry<Pair<String,String>, Map<String, Long>> r: XResRecExtension.instance().extractExpertiseResources(log).entrySet()){
			String rId = r.getKey().getFirst();
			for (Entry<String, Long> e: r.getValue().entrySet()){
				String eId = e.getKey();
				long value = e.getValue();
				result.addResourceExpertise(rId, eId, value);
			}
			
		}
		
		//Workload
		for(Entry<Pair<String, String>, Map<String, Long>> w: XResRecExtension.instance().extractWorkloads(log).entrySet()){
			String rId = w.getKey().getFirst();
			Map<String,Long> expertises = w.getValue();
			Long max = expertises.get(XResRecExtension.KEY_WORKLOAD_MAX);
			Long current = expertises.get(XResRecExtension.KEY_WORKLOAD_CURRENT);
			result.addWorkload(rId,max,current);
		}
		
		
		// Historical Information
		
		for (XTrace trace: log){
			for(XEvent event: trace){
				
				// For COMPLETE events
				if (XLifecycleExtension.instance().extractStandardTransition(event).equals(XLifecycleExtension.StandardModel.COMPLETE)){
					
					String characterizationId = XConceptExtension.instance().extractInstance(event);
					String resourceId = XOrganizationalExtension.instance().extractResource(event);
					result.addFrequency(characterizationId, resourceId);
					XAttributeDiscrete quality = (XAttributeDiscrete) event.getAttributes().get("quality");
					result.addQuality(characterizationId, resourceId, quality.getValue());
					XAttributeContinuous duration =  (XAttributeContinuous) event.getAttributes().get("duration");
					result.addPerformance(characterizationId, resourceId, duration.getValue());
					Double cost = XCostExtension.instance().extractTotal(event);
					result.addCost(characterizationId, resourceId, cost);
				}
			}
		}
		return result;
	}
}
