package org.processmining.resourcerecommendation.algorithms;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.processmining.resourcerecommendation.algorithms.bpa.AnswerElement;
import org.processmining.resourcerecommendation.algorithms.bpa.Database;
import org.processmining.resourcerecommendation.algorithms.bpa.MyInteger;
import org.processmining.resourcerecommendation.algorithms.bpa.TopkAlgorithms;
import org.processmining.resourcerecommendation.models.Dimension;
import org.processmining.resourcerecommendation.models.RecommendationElement;
import org.processmining.resourcerecommendation.models.Request;
import org.processmining.resourcerecommendation.models.Resource;
import org.processmining.resourcerecommendation.models.ResourceKnowledge;
import org.processmining.resourcerecommendation.models.ResourceRecommendation;
import org.processmining.resourcerecommendation.models.impl.ResourceProcessFactory;
import org.processmining.resourcerecommendation.parameters.RecommendResourcesParameters;

public class RecommendationAlgorithm {
	
	public ResourceRecommendation apply(ResourceKnowledge knowledge, RecommendResourcesParameters parameters){
		ResourceRecommendation result = null;
		int numRequests = parameters.getRequests().size();
		if (numRequests == 1){
			result = recommendOnDemand(knowledge, parameters);
		}
		else if (numRequests > 1){
			result = recommendBatch(knowledge, parameters);
		}
		return result;	
	}
	
	
	public ResourceRecommendation recommendOnDemand(ResourceKnowledge knowledge, RecommendResourcesParameters parameters){
		List<Resource> resources = knowledge.getResources();
		List<Dimension> dimensions = knowledge.getDimensions();
		Request request = parameters.getRequests().get(0); //Only one request
		Map<String, Double> weights = parameters.getWeights();
		int numRecommendations = parameters.getNumRecommendations();
		
		//Prepare the database for BPA2
		Database db = new Database(dimensions.size(), resources.size());
		for(int r=0; r<resources.size(); r++){
			for(int d=0; d<dimensions.size(); d++){
				String dimensionId = dimensions.get(d).getId();
				String characterizationId = request.getCharacterizationId();
				String resourceId = resources.get(r).getId();
				double metric = knowledge.getMetric(dimensionId, characterizationId, resourceId);
				double weight = weights.get(dimensionId);
				double score = metric * weight;
				db.lists[d].elements.get(r).dataID = r;
				db.lists[d].elements.get(r).score = score;
			}
		}
		for(int r=0; r<resources.size(); r++){
			Boolean feasible = true;	
			
			for(int d=0; d<dimensions.size(); d++) {
				String dimensionId = dimensions.get(d).getId();
				double weight = weights.get(dimensionId);
				
				if (weight!=0.0 && db.lists[d].elements.get(r).score == -10.0)
					feasible = false;
			}
			
			if (!feasible){
				for(int d=0; d<dimensions.size(); d++) {
					db.lists[d].elements.get(r).score = -10.0;
				}
			}
		}
		db.printDatabase();
		db.sortLists();
		db.printDatabase();
		
		//Execute BPA2
		TopkAlgorithms algorithms = new TopkAlgorithms(db);
		MyInteger stopPosition = new MyInteger();
		MyInteger directAccess = new MyInteger();
		MyInteger randomAccess = new MyInteger();
		AnswerElement [] answers = new AnswerElement[numRecommendations];
		algorithms.BPA2(numRecommendations, stopPosition, directAccess, randomAccess, answers);
		algorithms.printAnswers(answers, numRecommendations); 

		//Fill recommendation with results of BPA
		ResourceRecommendation recommendation = ResourceProcessFactory.createResourceRecommendation();
		List<RecommendationElement> ranking = new ArrayList<RecommendationElement>();
		for(int pos=0; pos < answers.length; pos++){
			Resource resource = resources.get(answers[pos].dataID);
			RecommendationElement elem = ResourceProcessFactory.createRecommendationElement(resource.getId(),resource.getDescription());
			double overallScore = answers[pos].overallScore;
			elem.setOverallScore(overallScore);
			
			for(int d=0; d<dimensions.size(); d++){
				String dimensionId = dimensions.get(d).getId();
				int dataPosition = db.dataPositions[answers[pos].dataID].positionsInLists[d];
	            double metric = db.lists[d].elements.get(dataPosition).score;
				elem.setScore(dimensionId, metric);
			}			
			ranking.add(elem);
		}
		recommendation.setRanking(request, ranking);
		return recommendation;
	}
	
	public ResourceRecommendation recommendBatch(ResourceKnowledge knowledge, RecommendResourcesParameters parameters){		
		List<Resource> resources = knowledge.getResources();
		List<Dimension> dimensions = knowledge.getDimensions();
		Map<String, Double> weights = parameters.getWeights();

		int numRecommendations = parameters.getNumRecommendations();
		int numRequests = parameters.getRequests().size();
		int numResources = resources.size();
		
		//Prepare the database for BPA2
		Database db = new Database(dimensions.size()*2, (int) Math.pow(numResources, numRequests));
		ArrayList<int []> cpList = generateCP(numRequests, numResources);
		mapPermute(knowledge, parameters, cpList, db);

		AnswerElement [] answers = new AnswerElement[numRecommendations];
		ResourceRecommendation recom = ResourceProcessFactory.createResourceRecommendation();

		boolean useBPA2 = true;
		
		if (useBPA2) {
			// db.printDatabase();
			db.sortLists();
			// db.printDatabase();

			//Execute BPA2
			TopkAlgorithms algorithms = new TopkAlgorithms(db);
			MyInteger stopPosition = new MyInteger();
			MyInteger directAccess = new MyInteger();
			MyInteger randomAccess = new MyInteger();
			algorithms.BPA2(numRecommendations, stopPosition, directAccess, randomAccess, answers);
			algorithms.printAnswers(answers, numRecommendations);
			
			//Fill recommendation with results of BPA
			for(int j=0; j<numRequests; j++){
				Request request = parameters.getRequests().get(j); //Multiple requests
				List<RecommendationElement> ranking = new ArrayList<RecommendationElement>();
				for(int pos=0; pos < answers.length; pos++){
					int best = answers[pos].dataID;
					int[] bestResources = cpList.get(best);
					
					Resource resource = resources.get(bestResources[j]);
					RecommendationElement elem = ResourceProcessFactory.createRecommendationElement(resource.getId(),resource.getDescription());
					double overallScore = answers[pos].overallScore;
					elem.setOverallScore(overallScore);
					
					for(int d=0; d<dimensions.size(); d++){
						String dimensionId = dimensions.get(d).getId();
						int dataPosition = db.dataPositions[answers[pos].dataID].positionsInLists[d];
	                    double metric = db.lists[d].elements.get(dataPosition).score;
						elem.setScore(dimensionId, metric);
					}
					ranking.add(elem);
				}
				recom.setRanking(request, ranking);
			}
		} else {
			// db is still in the initial order dataID == cp_i
			
			for(int cp_i=0; cp_i<numRecommendations; cp_i++){
				answers[cp_i] = new AnswerElement();
				int dataID = cp_i;

				double value = 0.0;
				for(int d=0; d<db.lists.length; d++) {
					value += db.lists[d].elements.get(dataID).score;
				}

				answers[cp_i].overallScore = value;
				answers[cp_i].dataID = dataID;

				int k = cp_i;
				for (int j=cp_i; j<cp_i; j++) {
					if (answers[k].overallScore > answers[j].overallScore) {
						int tmpDataID = answers[j].dataID;
						double tmpOverallScore = answers[j].overallScore;
						answers[j].dataID = answers[k].dataID;
						answers[j].overallScore = answers[k].overallScore;
						answers[k].dataID = tmpDataID;
						answers[k].overallScore = tmpOverallScore;
						k = j;
					}
				}
			}
			
		    System.out.println();
			for(int j=0;j<answers.length; j++) {
			      System.out.println(answers[j].dataID + " : " + answers[j].overallScore + " ,   ");
			}
			
			for(int cp_i=numRecommendations; cp_i<cpList.size(); cp_i++){
				int dataID = cp_i;

				double value = 0.0;
				for(int d=0; d<db.lists.length; d++) {
					value += db.lists[d].elements.get(dataID).score;
				}

				for (int j=0; j<numRecommendations; j++) {
					if (value > answers[j].overallScore) {
						int tmpDataID = answers[j].dataID;
						double tmpOverallScore = answers[j].overallScore;
						answers[j].dataID = dataID;
						answers[j].overallScore = value;
						dataID = tmpDataID;
						value = tmpOverallScore;
					}
				}
			}

		    System.out.println();
			for(int j=0;j<answers.length; j++) {
			      System.out.println(answers[j].dataID + " : " + answers[j].overallScore + " ,   ");
			}

			// Fill recommendation with results; db is still in the initial order dataID == cp_i
			for(int j=0; j<numRequests; j++){
				Request request = parameters.getRequests().get(j); //Multiple requests
				List<RecommendationElement> ranking = new ArrayList<RecommendationElement>();
				for(int pos=0; pos < answers.length; pos++){
					int best = answers[pos].dataID;
					int[] bestResources = cpList.get(best);
					
					Resource resource = resources.get(bestResources[j]);
					RecommendationElement elem = ResourceProcessFactory.createRecommendationElement(resource.getId(),resource.getDescription());
					double overallScore = answers[pos].overallScore;
					elem.setOverallScore(overallScore);
					
					for(int d=0; d<dimensions.size(); d++){
						String dimensionId = dimensions.get(d).getId();
	                    double metric = db.lists[d].elements.get(answers[pos].dataID).score;
						elem.setScore(dimensionId, metric);
					}
					ranking.add(elem);
				}
				recom.setRanking(request, ranking);
			}

		}

		return recom;
	}
	
	public ArrayList<int []> generateCP(int numRequests, int numResources){
		int[] lengths = new int[numRequests];
		for (int req=0; req<numRequests; req++){
			lengths[req] = numResources;			
		}
		CartesianProduct cp = new CartesianProduct (lengths);

		ArrayList<int []> cpList = new ArrayList<int []>();
		for(int[] indices: cp){
			cpList.add(indices);
		}
		return cpList;
	}
	
	public void mapPermute(ResourceKnowledge knowledge, RecommendResourcesParameters parameters, ArrayList<int []> cpList, Database db){
		List<Resource> resources = knowledge.getResources();
		List<Dimension> dimensions = knowledge.getDimensions();
		Map<String, Double> weights = parameters.getWeights();

		// These values could be given as parameters
		double w_avg = 0.9;
		double w_stdev = 0.1;

		int numRequests = parameters.getRequests().size();
		int numResources = resources.size();
		
		String[] allResourceIds = new String[numResources];
		for(int r=0; r<numResources; r++)
			allResourceIds[r] = resources.get(r).getId();
		
		for(int d=0; d<dimensions.size(); d++) {
			String dimensionId = dimensions.get(d).getId();
			double weight = weights.get(dimensionId);
	
			String[] characterizationIds = new String[numRequests];
			for(int req=0; req<numRequests; req++){
				Request request = parameters.getRequests().get(req); //Multiple requests
				characterizationIds[req] = request.getCharacterizationId();
			}
			
			int d_stdev = d+dimensions.size();
			
			for(int cp_i = 0; cp_i < cpList.size(); cp_i++){
				int[] indices = cpList.get(cp_i);
				String[] resourceIds = new String[numRequests];
				for(int req=0; req<numRequests; req++){
					resourceIds[req]= allResourceIds[indices[req]];
				}
				
				Boolean feasible = true;
				
				double sum = 0;
				ArrayList<Double> metricByResources = knowledge.getMetric(dimensionId, characterizationIds, resourceIds);
				for(int req=0; req<numRequests; req++){
					if (metricByResources.get(req) == -10.0){
						feasible = false;
					}else {
						sum +=  metricByResources.get(req);
					}
				}
				if (feasible){
					double avg = sum/numRequests;
					double sumV = 0;
					for(int req=0; req<numRequests; req++)
						sumV += Math.pow((metricByResources.get(req) - avg),2);
					double stdev = 1- Math.sqrt(sumV/numRequests);
					db.lists[d].elements.get(cp_i).dataID = cp_i;
					db.lists[d].elements.get(cp_i).score = avg*w_avg*weight;
					db.lists[d_stdev].elements.get(cp_i).dataID = cp_i;
					db.lists[d_stdev].elements.get(cp_i).score = stdev*w_stdev*weight;
				}else{
					db.lists[d].elements.get(cp_i).dataID = cp_i;
					db.lists[d].elements.get(cp_i).score = -10.0;
					db.lists[d_stdev].elements.get(cp_i).dataID = cp_i;
					db.lists[d_stdev].elements.get(cp_i).score = -10.0;
				}
			}
		}
		for(int cp_i = 0; cp_i < cpList.size(); cp_i++){
			Boolean feasible = true;	
			
			for(int d=0; d<dimensions.size(); d++) {
				String dimensionId = dimensions.get(d).getId();
				double weight = weights.get(dimensionId);
				
				if (weight!=0.0 && db.lists[d].elements.get(cp_i).score == -10.0)
					feasible = false;
			}
			
			if (!feasible){
				for(int d=0; d<dimensions.size(); d++) {
					db.lists[d].elements.get(cp_i).score = -10.0;
					int d_stdev = d+dimensions.size();
					db.lists[d_stdev].elements.get(cp_i).score = -10.0;
				}
			}
		}
	}
}
