package org.processmining.resourcerecommendation.plugins;

import java.util.Collection;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.resourcerecommendation.algorithms.GenerateResourceKnowledgeAlgorithm;
import org.processmining.resourcerecommendation.connections.GenerateResourceKnowledgeConnection;
import org.processmining.resourcerecommendation.dialogs.GenerateResourceKnowledgeDialog;
import org.processmining.resourcerecommendation.help.GenerateResourceKnowledgeHelp;
import org.processmining.resourcerecommendation.models.ResourceKnowledge;
import org.processmining.resourcerecommendation.parameters.GenerateResourceKnowledgeParameters;

@Plugin(
		name = "Generate Resource Knowledge", 
		parameterLabels = { "Log", "Settings" },  
		returnLabels = { "Resource Knowledge"},
		returnTypes = { ResourceKnowledge.class},
		categories = {PluginCategory.Analytics, PluginCategory.Enhancement},
		keywords = {"Resource Allocation", "Recommendation Systems", "Organizational Perspective"},
		help = GenerateResourceKnowledgeHelp.PLUGIN_DESCRIPTION,
		userAccessible = true
		)
public class GenerateResourceKnowledgePlugin extends GenerateResourceKnowledgeAlgorithm {


	@PluginVariant(variantLabel = "Generate Resource Knowledge, parameters", requiredParameterLabels = {0, 1})
	public ResourceKnowledge run(PluginContext context, XLog log, GenerateResourceKnowledgeParameters parameters) {
		// Apply the algorithm depending on whether a connection already exists.
	    return runConnections(context, log, parameters);
	}
	
	
	@UITopiaVariant(
			affiliation = "Pontificia Universidad Catolica de Chile", 
			author = "Michael Arias, et al.", 
			email = "m.arias"+ (char) 0x40 + "uc.cl",
			uiLabel = UITopiaVariant.USEPLUGIN,
			pack = "ResourceRecommendation"
			)
	@PluginVariant(variantLabel = "Generate Resource Knowledge, default", requiredParameterLabels = { 0 })
	public ResourceKnowledge runDefault(PluginContext context, XLog log) {
		// Get the default parameters.
		GenerateResourceKnowledgeParameters parameters = new GenerateResourceKnowledgeParameters(log);
		// Apply the algorithm depending on whether a connection already exists.
	    return runConnections(context, log, parameters);
	}
	

	@UITopiaVariant(
			affiliation = "Pontificia Universidad Catolica de Chile", 
			author = "Michael Arias, et al.", 
			email = "m.arias"+ (char) 0x40 + "uc.cl",
			uiLabel = UITopiaVariant.USEPLUGIN,
			pack = "ResourceRecommendation"
			)
	@PluginVariant(variantLabel = "Generate Resource Knowledge, dialog", requiredParameterLabels = {0})
	public ResourceKnowledge runUI(UIPluginContext context, XLog log) {
		// Get the default parameters.
		GenerateResourceKnowledgeParameters parameters = new GenerateResourceKnowledgeParameters(log);
	    // Get a dialog for this parameters.
		GenerateResourceKnowledgeDialog dialog = new GenerateResourceKnowledgeDialog(context, log, parameters);
		int n = 0;
		int nPanels = dialog.getNumPanels();
		String[] title = dialog.getTitles();
		InteractionResult result = InteractionResult.NEXT;
 		while (result != InteractionResult.FINISHED) {
 			result = context.showWizard(title[n], n == 0, n == nPanels-1, dialog.getPanel(n));
 			if (result == InteractionResult.NEXT) {
 				n++;
 			} else if (result == InteractionResult.PREV) {
 				n--;
 			} else if (result != InteractionResult.FINISHED) {
 				return null;
 			}
 		}
 		
 		return run(context, log, parameters);	 		
	}	
	
	

	private ResourceKnowledge runConnections(PluginContext context, XLog log, GenerateResourceKnowledgeParameters parameters) {
		if (parameters.isTryConnections()) {
			// Try to found a connection that matches the inputs and the parameters.
			Collection<GenerateResourceKnowledgeConnection> connections;
			try {
				connections = context.getConnectionManager().getConnections(
						GenerateResourceKnowledgeConnection.class, context, log);
				for (GenerateResourceKnowledgeConnection connection : connections) {
					if (connection.getObjectWithRole(GenerateResourceKnowledgeConnection.LOG)
							.equals(log) && connection.getParameters().equals(parameters)) {
						// Found a match. Return the associated output as result of the algorithm.
						return connection
								.getObjectWithRole(GenerateResourceKnowledgeConnection.KNOWLEDGE);
					}
				}
			} catch (ConnectionCannotBeObtained e) {
			}
		}
		// No connection found. Apply the algorithm to compute a fresh output result.
		ResourceKnowledge output = apply(context, log, parameters);
		if (parameters.isTryConnections()) {
			// Store a connection containing the inputs, output, and parameters.
			context.getConnectionManager().addConnection(
					new GenerateResourceKnowledgeConnection(log, output, parameters));
		}
		// Return the output.
		return output;
	}

}
