package org.processmining.resourcerecommendation.plugins;

import java.util.Collection;

import lpsolve.LpSolveException;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.resourcerecommendation.algorithms.RecommendResourcesAlgorithm;
import org.processmining.resourcerecommendation.connections.RecommendResourcesConnection;
import org.processmining.resourcerecommendation.dialogs.RecommendResourcesDialog;
import org.processmining.resourcerecommendation.help.RecommendResourcesHelp;
import org.processmining.resourcerecommendation.models.ResourceKnowledge;
import org.processmining.resourcerecommendation.models.ResourceRecommendation;
import org.processmining.resourcerecommendation.parameters.RecommendResourcesParameters;

@Plugin(
		name = "Recommend Resources", 
		parameterLabels = { "Resource Knowledge",  "Settings" },  
		returnLabels = { "Resource Recommendation"},
		returnTypes = { ResourceRecommendation.class},
		categories = {PluginCategory.Analytics, PluginCategory.Enhancement},
		keywords = {"Resource Allocation", "Recommendation Systems", "Organizational Perspective"},
		help = RecommendResourcesHelp.PLUGIN_DESCRIPTION,
		userAccessible = true
		)
public class RecommendResourcesPlugin extends RecommendResourcesAlgorithm {


	@PluginVariant(variantLabel = "Recommend Resources, parameters", requiredParameterLabels = {0, 1})
	public ResourceRecommendation run(PluginContext context, ResourceKnowledge knowledge, RecommendResourcesParameters parameters) throws LpSolveException {
		// Apply the algorithm depending on whether a connection already exists.
	    return runConnections(context, knowledge, parameters);
	}
	
	
	@UITopiaVariant(
			affiliation = "Pontificia Universidad Catolica de Chile", 
			author = "Michael Arias, et al.", 
			email = "m.arias"+ (char) 0x40 + "uc.cl",
			uiLabel = UITopiaVariant.USEPLUGIN,
			pack = "ResourceRecommendation"
			)
	@PluginVariant(variantLabel = "Recommend Resources, default", requiredParameterLabels = { 0 })
	public ResourceRecommendation runDefault(PluginContext context, ResourceKnowledge knowledge) throws LpSolveException {
		// Get the default parameters.
		RecommendResourcesParameters parameters = new RecommendResourcesParameters(knowledge);
		// Apply the algorithm depending on whether a connection already exists.
	    return runConnections(context, knowledge, parameters);
	}
	

	@UITopiaVariant(
			affiliation = "Pontificia Universidad Catolica de Chile", 
			author = "Michael Arias, et al.", 
			email = "m.arias"+ (char) 0x40 + "uc.cl",
			uiLabel = UITopiaVariant.USEPLUGIN,
			pack = "ResourceRecommendation"
			)
	@PluginVariant(variantLabel = "Recommend Resources, dialog", requiredParameterLabels = {0})
	public ResourceRecommendation runUI(UIPluginContext context, ResourceKnowledge knowledge) throws LpSolveException {
		// Get the default parameters.
		RecommendResourcesParameters parameters = new RecommendResourcesParameters(knowledge);
	    // Get a dialog for this parameters.
		RecommendResourcesDialog dialog = new RecommendResourcesDialog(context, knowledge, parameters);
		int n = 0;
		int nPanels = dialog.getNumPanels();
		String[] title = dialog.getTitles();
		InteractionResult result = InteractionResult.NEXT;
		while (result != InteractionResult.FINISHED) {
			result = context.showWizard(title[n], n == 0, n == nPanels-1, dialog.getPanel(n));
			if (result == InteractionResult.NEXT) {
				n++;
			} else if (result == InteractionResult.PREV) {
				n--;
			} else if (result != InteractionResult.FINISHED) {
				return null;
			}
		}
		
		//TODO Remove
		System.out.println(parameters.getAlgorithm());
		
		return run(context, knowledge, parameters);
	}	

	private ResourceRecommendation runConnections(PluginContext context, ResourceKnowledge knowledge, RecommendResourcesParameters parameters) throws LpSolveException {
		if (parameters.isTryConnections()) {
			// Try to found a connection that matches the inputs and the parameters.
			Collection<RecommendResourcesConnection> connections;
			try {
				connections = context.getConnectionManager().getConnections(
						RecommendResourcesConnection.class, context, knowledge);
				for (RecommendResourcesConnection connection : connections) {
					if (connection.getObjectWithRole(RecommendResourcesConnection.KNOWLEDGE)
							.equals(knowledge) && connection.getParameters().equals(parameters)) {
						// Found a match. Return the associated output as result of the algorithm.
						return connection
								.getObjectWithRole(RecommendResourcesConnection.RECOMMENDATION);
					}
				}
			} catch (ConnectionCannotBeObtained e) {
			}
		}
		// No connection found. Apply the algorithm to compute a fresh output result.
		ResourceRecommendation output = apply(context, knowledge, parameters);
		if (parameters.isTryConnections()) {
			// Store a connection containing the inputs, output, and parameters.
			context.getConnectionManager().addConnection(
					new RecommendResourcesConnection(knowledge, output, parameters));
		}
		// Return the output.
		return output;
	}

}
