// Copyright (C) 2014 by Jorge Munoz-Gama. All rights reserved. 
// Released under the terms of the Lesser GNU General Public License (L-GPL) version 2 or later.

package org.processmining.resourcerecommendation.utils.resrecxes;

import java.net.URI;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XGlobalAttributeNameMap;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeContainer;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.util.Pair;

public class XResRecExtension extends XExtension{

	private static final long serialVersionUID = 6132894478608626123L;
	
	/**
	 * Unique URI of this extension.
	 */
	public static final URI EXTENSION_URI = URI
			.create("http://www.processmininguc.org/xes/resrec.xesext");
	
	/**
	 * Keys for the attributes.
	 */
	public static final String KEY_RESOURCES = "resrec:resources";
	public static final String KEY_WORKLOADS = "resrec:workloads";
	public static final String KEY_WORKLOAD_MAX = "max";
	public static final String KEY_WORKLOAD_CURRENT = "current";
	public static final String KEY_UNITS = "resrec:units";
	public static final String KEY_FACTORS = "resrec:factors";
	public static final String KEY_TYPOLOGIES = "resrec:typologies";
	public static final String KEY_CHARACTERIZATIONS = "resrec:characterizations";
	public static final String KEY_CHARACTERIZATION_UNIT = "unit";
	public static final String KEY_CHARACTERIZATION_TYPOLOGY = "typology";
	public static final String KEY_EXPERTISES = "resrec:expertises";
	public static final String KEY_EXPERTISE_RESOURCES = "resrec:expertise_resources";
	public static final String KEY_EXPERTISE_REQUIRED = "resrec:expertise_required";
	
	/**
	 * Attribute prototypes
	 */
	public static XAttributeContainer ATTR_RESOURCES;
	public static XAttributeContainer ATTR_WORKLOADS;
	public static XAttributeContainer ATTR_UNITS;
	public static XAttributeContainer ATTR_FACTORS;
	public static XAttributeContainer ATTR_TYPOLOGIES;
	public static XAttributeContainer ATTR_CHARACTERIZATIONS;
	public static XAttributeContainer ATTR_EXPERTISES;
	public static XAttributeContainer ATTR_EXPERTISE_RESOURCES;
	public static XAttributeContainer ATTR_EXPERTISE_REQUIRED;
	
	
	/**
	 * Singleton instance of this extension.
	 */
	private transient static XResRecExtension singleton = new XResRecExtension();

	/**
	 * Provides access to the singleton instance.
	 * 
	 * @return Singleton extension.
	 */
	public static XResRecExtension instance() {
		return singleton;
	}

	private Object readResolve() {
		return singleton;
	}

	/**
	 * Private constructor
	 */
	private XResRecExtension() {
		super("ResRec", "resrec", EXTENSION_URI);
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		
		ATTR_RESOURCES = factory.createAttributeContainer(KEY_RESOURCES, this);
		ATTR_WORKLOADS = factory.createAttributeContainer(KEY_WORKLOADS, this);
		ATTR_UNITS = factory.createAttributeContainer(KEY_UNITS, this);
		ATTR_FACTORS = factory.createAttributeContainer(KEY_FACTORS, this);
		ATTR_TYPOLOGIES = factory.createAttributeContainer(KEY_TYPOLOGIES, this);
		ATTR_CHARACTERIZATIONS = factory.createAttributeContainer(KEY_CHARACTERIZATIONS, this);
		ATTR_EXPERTISES = factory.createAttributeContainer(KEY_EXPERTISES, this);
		ATTR_EXPERTISE_RESOURCES = factory.createAttributeContainer(KEY_EXPERTISE_RESOURCES, this);
		ATTR_EXPERTISE_REQUIRED = factory.createAttributeContainer(KEY_EXPERTISE_REQUIRED, this);
		
		this.logAttributes.add((XAttribute) ATTR_RESOURCES.clone());
		this.logAttributes.add((XAttribute) ATTR_WORKLOADS.clone());
		this.logAttributes.add((XAttribute) ATTR_UNITS.clone());
		this.logAttributes.add((XAttribute) ATTR_FACTORS.clone());
		this.logAttributes.add((XAttribute) ATTR_TYPOLOGIES.clone());
		this.logAttributes.add((XAttribute) ATTR_CHARACTERIZATIONS.clone());
		this.logAttributes.add((XAttribute) ATTR_EXPERTISES.clone());
		this.logAttributes.add((XAttribute) ATTR_EXPERTISE_RESOURCES.clone());
		this.logAttributes.add((XAttribute) ATTR_EXPERTISE_REQUIRED.clone());
		
		//register aliases
		XGlobalAttributeNameMap.instance().registerMapping(
				XGlobalAttributeNameMap.MAPPING_ENGLISH, KEY_RESOURCES, "Resources");
		XGlobalAttributeNameMap.instance().registerMapping(
				XGlobalAttributeNameMap.MAPPING_ENGLISH, KEY_WORKLOADS, "Workloads");
		XGlobalAttributeNameMap.instance().registerMapping(
				XGlobalAttributeNameMap.MAPPING_ENGLISH, KEY_UNITS, "Units");
		XGlobalAttributeNameMap.instance().registerMapping(
				XGlobalAttributeNameMap.MAPPING_ENGLISH, KEY_FACTORS, "Factors");
		XGlobalAttributeNameMap.instance().registerMapping(
				XGlobalAttributeNameMap.MAPPING_ENGLISH, KEY_TYPOLOGIES, "Typologies");
		XGlobalAttributeNameMap.instance().registerMapping(
				XGlobalAttributeNameMap.MAPPING_ENGLISH, KEY_CHARACTERIZATIONS, "Characterizations");
		XGlobalAttributeNameMap.instance().registerMapping(
				XGlobalAttributeNameMap.MAPPING_ENGLISH, KEY_EXPERTISES, "Expertises");
		XGlobalAttributeNameMap.instance().registerMapping(
				XGlobalAttributeNameMap.MAPPING_ENGLISH, KEY_EXPERTISE_RESOURCES, "Expertise Resources");
		XGlobalAttributeNameMap.instance().registerMapping(
				XGlobalAttributeNameMap.MAPPING_ENGLISH, KEY_EXPERTISE_REQUIRED, "Expertise Required");
	}
	
	public Map<String,String> extractResources(XLog log){
		XAttribute attribute = log.getAttributes().get(KEY_RESOURCES);
		if (attribute == null){
			return null;
		}
		else{
			XAttributeContainer container = (XAttributeContainer) attribute;
			Map<String, String> result = new LinkedHashMap<String,String>();
			for(XAttribute resource: container.getAttributes().values()){
				XAttributeLiteral resourceLit = (XAttributeLiteral) resource;
				result.put(resourceLit.getKey(), resourceLit.getValue());
			}
			return result;
		}
	}
	
	public void assignResources(XLog log, Map<String,String> map){
		if(map != null && map.size() > 0 ){
			XAttributeContainer attr = (XAttributeContainer) ATTR_RESOURCES.clone();
			XFactory factory = XFactoryRegistry.instance().currentDefault();
			for(Entry<String, String> entry: map.entrySet()){ 
				XAttributeLiteral resource = factory.createAttributeLiteral(entry.getKey(), entry.getValue(), this);
				attr.getAttributes().put(entry.getKey(), resource);
			}
			
			log.getAttributes().put(KEY_RESOURCES, attr);
		}
	}
	
	public Map<Pair<String,String>,Map<String,Long>> extractWorkloads(XLog log){
		XAttribute attribute = log.getAttributes().get(KEY_WORKLOADS);
		if (attribute == null){
			return null;
		}
		else{
			XAttributeContainer container = (XAttributeContainer) attribute;
			Map<Pair<String,String>,Map<String,Long>> result = new LinkedHashMap<Pair<String,String>,Map<String,Long>>();
			for (XAttribute resource: container.getAttributes().values()){
				XAttributeLiteral resourceLit = (XAttributeLiteral) resource;
				String resourceId = resourceLit.getKey();
				String resourceDescription = resourceLit.getValue();
				Pair<String,String> keys = new Pair<String,String>(resourceId,resourceDescription);
				Map<String,Long> values = new LinkedHashMap<String,Long>();
				Long maxValue = ((XAttributeDiscrete) resourceLit.getAttributes().get(KEY_WORKLOAD_MAX)).getValue();
				values.put(KEY_WORKLOAD_MAX, maxValue);
				Long currentValue = ((XAttributeDiscrete) resourceLit.getAttributes().get(KEY_WORKLOAD_CURRENT)).getValue();
				values.put(KEY_WORKLOAD_CURRENT, currentValue);
				result.put(keys, values);
			}
			return result;
		}
	}
	
	public void assignWorkloads(XLog log, Map<Pair<String,String>,Map<String,Long>> map){
		if(map != null && map.size() > 0){
			XAttributeContainer attr = (XAttributeContainer) ATTR_WORKLOADS.clone();
			XFactory factory = XFactoryRegistry.instance().currentDefault();
			for(Entry<Pair<String,String>, Map<String,Long>> entry: map.entrySet()){
				String resourceId = entry.getKey().getFirst();
				String resourceDescription = entry.getKey().getSecond();
				Long maxValue = entry.getValue().get(KEY_WORKLOAD_MAX);
				Long currentValue = entry.getValue().get(KEY_WORKLOAD_CURRENT);
				
				
				XAttributeLiteral workLit = factory.createAttributeLiteral(resourceId, resourceDescription, this);
				XAttributeDiscrete maxLit = factory.createAttributeDiscrete(KEY_WORKLOAD_MAX, maxValue, this);
				XAttributeDiscrete currentLit = factory.createAttributeDiscrete(KEY_WORKLOAD_CURRENT, currentValue, this);
				workLit.getAttributes().put(KEY_WORKLOAD_MAX, maxLit);
				workLit.getAttributes().put(KEY_WORKLOAD_CURRENT, currentLit);				
				attr.getAttributes().put(resourceId, workLit);	
			}
			log.getAttributes().put(KEY_WORKLOADS, attr);
		}
	}
	
	public Map<String, String> extractUnits(XLog log){
		XAttribute attribute = log.getAttributes().get(KEY_UNITS);
		if(attribute == null){
			return null;
		}
		else{
			XAttributeContainer container = (XAttributeContainer) attribute;
			Map<String, String> result = new LinkedHashMap<String,String>();
			for(XAttribute unit: container.getAttributes().values()){
				XAttributeLiteral unitLit = (XAttributeLiteral) unit;
				result.put(unitLit.getKey(), unitLit.getValue());
			}
			return result;
		}
	}
	
	public void assignUnits(XLog log, Map<String,String> map){
		if(map != null && map.size() > 0 ){
			XAttributeContainer attr = (XAttributeContainer) ATTR_UNITS.clone();
			XFactory factory = XFactoryRegistry.instance().currentDefault();
			for(Entry<String, String> entry: map.entrySet()){ 
				XAttributeLiteral unit = factory.createAttributeLiteral(entry.getKey(), entry.getValue(), this);
				attr.getAttributes().put(entry.getKey(), unit);
			}
			
			log.getAttributes().put(KEY_UNITS, attr);
		}
	}
	
	public Map<Pair<String,String>,Map<String,String>> extractFactors(XLog log){
		XAttribute attribute = log.getAttributes().get(KEY_FACTORS);
		if (attribute == null){
			return null;
		}
		else{
			XAttributeContainer container = (XAttributeContainer) attribute;
			Map<Pair<String,String>,Map<String,String>> result = new LinkedHashMap<Pair<String,String>,Map<String,String>>();
			for(XAttribute factor: container.getAttributes().values()){
				XAttributeLiteral factorLit = (XAttributeLiteral) factor;
				String factorId = factorLit.getKey();
				String factorDescription = factorLit.getValue();
				Pair<String,String> pair = new Pair<String,String>(factorId, factorDescription);
				Map<String,String> options = new LinkedHashMap<String, String>();
				for(XAttribute option: factor.getAttributes().values()){
					XAttributeLiteral optionLit = (XAttributeLiteral) option;
					String optionId = optionLit.getKey();
					String optionDescription = optionLit.getValue();
					options.put(optionId, optionDescription);
				}
				result.put(pair, options);
			}
			return result;
		}
	}
	
	public void assignFactors(XLog log, Map<Pair<String,String>,Map<String,String>> map){
		if(map != null && map.size() > 0){
			XAttributeContainer attr = (XAttributeContainer) ATTR_FACTORS.clone();
			XFactory factory = XFactoryRegistry.instance().currentDefault();
			for(Entry<Pair<String,String>,Map<String,String>> entry: map.entrySet()){
				String factorId = entry.getKey().getFirst();
				String factorDescription = entry.getKey().getSecond();
				XAttributeLiteral factorLit = factory.createAttributeLiteral(factorId, factorDescription, this);
				
				for(Entry<String,String> option: entry.getValue().entrySet()){
					String optionId = option.getKey();
					String optionDescription = option.getValue();
					XAttributeLiteral optionLit = factory.createAttributeLiteral(optionId, optionDescription, this);
					factorLit.getAttributes().put(optionId, optionLit);
				}				
				attr.getAttributes().put(factorId, factorLit);	
			}
			log.getAttributes().put(KEY_FACTORS, attr);
		}
	}
	
	public Map<Pair<String,String>,Map<String,String>> extractTypologies(XLog log){
		XAttribute attribute = log.getAttributes().get(KEY_TYPOLOGIES);
		if (attribute == null){
			return null;
		}
		else{
			XAttributeContainer container = (XAttributeContainer) attribute;
			Map<Pair<String,String>,Map<String,String>> result = new LinkedHashMap<Pair<String,String>,Map<String,String>>();
			for(XAttribute typology: container.getAttributes().values()){
				XAttributeLiteral typologyLit = (XAttributeLiteral) typology;
				String typologyId = typologyLit.getKey();
				String typologyDescription = typologyLit.getValue();
				Pair<String,String> pair = new Pair<String,String>(typologyId, typologyDescription);
				Map<String,String> options = new LinkedHashMap<String, String>();
				for(XAttribute option: typology.getAttributes().values()){
					XAttributeLiteral optionLit = (XAttributeLiteral) option;
					String optionId = optionLit.getKey();
					String optionDescription = optionLit.getValue();
					options.put(optionId, optionDescription);
				}
				result.put(pair, options);
			}
			return result;
		}
	}
	
	public void assignTypologies(XLog log, Map<Pair<String,String>,Map<String,String>> map){
		if(map != null && map.size() > 0){
			XAttributeContainer attr = (XAttributeContainer) ATTR_TYPOLOGIES.clone();
			XFactory factory = XFactoryRegistry.instance().currentDefault();
			for(Entry<Pair<String,String>,Map<String,String>> entry: map.entrySet()){
				String typologyId = entry.getKey().getFirst();
				String typologyDescription = entry.getKey().getSecond();
				XAttributeLiteral typologyLit = factory.createAttributeLiteral(typologyId, typologyDescription, this);
				
				for(Entry<String,String> option: entry.getValue().entrySet()){
					String optionId = option.getKey();
					String optionDescription = option.getValue();
					XAttributeLiteral optionLit = factory.createAttributeLiteral(optionId, optionDescription, this);
					typologyLit.getAttributes().put(optionId, optionLit);
				}				
				attr.getAttributes().put(typologyId, typologyLit);	
			}
			log.getAttributes().put(KEY_TYPOLOGIES, attr);
		}
	}
	
	public Map<Pair<String,String>,Pair<String,String>> extractCharacterizations(XLog log){
		XAttribute attribute = log.getAttributes().get(KEY_CHARACTERIZATIONS);
		if (attribute == null){
			return null;
		}
		else{
			XAttributeContainer container = (XAttributeContainer) attribute;
			Map<Pair<String,String>,Pair<String,String>> result = new LinkedHashMap<Pair<String,String>,Pair<String,String>>();
			for(XAttribute characterization: container.getAttributes().values()){
				XAttributeLiteral characterizationLit = (XAttributeLiteral) characterization;
				String characterizationId = characterizationLit.getKey();
				String characterizationDescription = characterizationLit.getValue();
				String unitId = ((XAttributeLiteral) characterizationLit.getAttributes().get(KEY_CHARACTERIZATION_UNIT)).getValue();
				String typologyId = ((XAttributeLiteral) characterizationLit.getAttributes().get(KEY_CHARACTERIZATION_TYPOLOGY)).getValue();
				Pair<String,String> pair1 = new Pair<String,String>(characterizationId, characterizationDescription);
				Pair<String,String> pair2 = new Pair<String,String>(unitId, typologyId);
				result.put(pair1, pair2);
			}
			return result;
		}
	}
	
	public void assignCharacterizations(XLog log, Map<Pair<String,String>,Pair<String,String>> map){
		if(map != null && map.size() > 0){
			XAttributeContainer attr = (XAttributeContainer) ATTR_CHARACTERIZATIONS.clone();
			XFactory factory = XFactoryRegistry.instance().currentDefault();
			for(Entry<Pair<String,String>, Pair<String,String>> entry: map.entrySet()){
				String characterizationId = entry.getKey().getFirst();
				String characterizationDescription = entry.getKey().getSecond();
				String unitId = entry.getValue().getFirst();
				String typologyId =  entry.getValue().getSecond();
				
				XAttributeLiteral charLit = factory.createAttributeLiteral(characterizationId, characterizationDescription, this);
				XAttributeLiteral unitLit = factory.createAttributeLiteral(KEY_CHARACTERIZATION_UNIT, unitId, this);
				XAttributeLiteral typologyLit = factory.createAttributeLiteral(KEY_CHARACTERIZATION_TYPOLOGY, typologyId, this);
				charLit.getAttributes().put(KEY_CHARACTERIZATION_UNIT, unitLit);
				charLit.getAttributes().put(KEY_CHARACTERIZATION_TYPOLOGY, typologyLit);
				
				attr.getAttributes().put(characterizationId, charLit);	
			}
			log.getAttributes().put(KEY_CHARACTERIZATIONS, attr);
		}
	}
	
	public Map<String, String> extractExpertises(XLog log){
		XAttribute attribute = log.getAttributes().get(KEY_EXPERTISES);
		if(attribute == null){
			return null;
		}
		else{
			XAttributeContainer container = (XAttributeContainer) attribute;
			Map<String, String> result = new LinkedHashMap<String,String>();
			for(XAttribute expertise: container.getAttributes().values()){
				XAttributeLiteral expertiseLit = (XAttributeLiteral) expertise;
				result.put(expertiseLit.getKey(), expertiseLit.getValue());
			}
			return result;
		}
	}
	
	public void assignExpertises(XLog log, Map<String,String> map){
		if(map != null && map.size() > 0 ){
			XAttributeContainer attr = (XAttributeContainer) ATTR_EXPERTISES.clone();
			XFactory factory = XFactoryRegistry.instance().currentDefault();
			for(Entry<String, String> entry: map.entrySet()){ 
				XAttributeLiteral expertise = factory.createAttributeLiteral(entry.getKey(), entry.getValue(), this);
				attr.getAttributes().put(entry.getKey(), expertise);
			}
			
			log.getAttributes().put(KEY_EXPERTISES, attr);
		}
	}
	
	public Map<Pair<String,String>,Map<String,Long>> extractExpertiseResources(XLog log){
		XAttribute attribute = log.getAttributes().get(KEY_EXPERTISE_RESOURCES);
		if (attribute == null){
			return null;
		}
		else{
			XAttributeContainer container = (XAttributeContainer) attribute;
			Map<Pair<String,String>,Map<String,Long>> result = new LinkedHashMap<Pair<String,String>,Map<String,Long>>();
			for(XAttribute resource: container.getAttributes().values()){
				XAttributeLiteral resourceLit = (XAttributeLiteral) resource;
				String resourceId = resourceLit.getKey();
				String resourceDescription = resourceLit.getValue();
				Pair<String,String> pair = new Pair<String,String>(resourceId, resourceDescription);
				Map<String,Long> skills = new LinkedHashMap<String, Long>();
				for(XAttribute skill: resource.getAttributes().values()){
					XAttributeDiscrete skillDis = (XAttributeDiscrete) skill;
					String skillId = skillDis.getKey();
					Long skillValue = skillDis.getValue();
					skills.put(skillId, skillValue);
				}
				result.put(pair, skills);
			}
			return result;
		}
	}
	
	
	public void assignExpertiseResources(XLog log, Map<Pair<String,String>,Map<String,Long>> map){
		if(map != null && map.size() > 0){
			XAttributeContainer attr = (XAttributeContainer) ATTR_EXPERTISE_RESOURCES.clone();
			XFactory factory = XFactoryRegistry.instance().currentDefault();
			for(Entry<Pair<String,String>,Map<String,Long>> entry: map.entrySet()){
				String resourceId = entry.getKey().getFirst();
				String resourceDescription = entry.getKey().getSecond();
				XAttributeLiteral resourceLit = factory.createAttributeLiteral(resourceId, resourceDescription, this);
				
				for(Entry<String,Long> skill: entry.getValue().entrySet()){
					String skillId = skill.getKey();
					Long skillValue = skill.getValue();
					XAttributeDiscrete skillDis = factory.createAttributeDiscrete(skillId, skillValue, this);
					resourceLit.getAttributes().put(skillId, skillDis);
				}				
				attr.getAttributes().put(resourceId, resourceLit);	
			}
			log.getAttributes().put(KEY_EXPERTISE_RESOURCES, attr);
		}
	}
	
	
	
	
	public Map<Pair<String,String>,Map<String,Long>> extractExpertiseRequired(XLog log){
		XAttribute attribute = log.getAttributes().get(KEY_EXPERTISE_REQUIRED);
		if (attribute == null){
			return null;
		}
		else{
			XAttributeContainer container = (XAttributeContainer) attribute;
			Map<Pair<String,String>,Map<String,Long>> result = new LinkedHashMap<Pair<String,String>,Map<String,Long>>();
			for(XAttribute characterization: container.getAttributes().values()){
				XAttributeLiteral characterizationLit = (XAttributeLiteral) characterization;
				String characterizationId = characterizationLit.getKey();
				String characterizationDescription = characterizationLit.getValue();
				Pair<String,String> pair = new Pair<String,String>(characterizationId, characterizationDescription);
				Map<String,Long> skills = new LinkedHashMap<String, Long>();
				for(XAttribute skill: characterization.getAttributes().values()){
					XAttributeDiscrete skillDis = (XAttributeDiscrete) skill;
					String skillId = skillDis.getKey();
					Long skillValue = skillDis.getValue();
					skills.put(skillId, skillValue);
				}
				result.put(pair, skills);
			}
			return result;
		}
	}
	
	public void assignExpertiseRequired(XLog log, Map<Pair<String,String>,Map<String,Long>> map){
		if(map != null && map.size() > 0){
			XAttributeContainer attr = (XAttributeContainer) ATTR_EXPERTISE_REQUIRED.clone();
			XFactory factory = XFactoryRegistry.instance().currentDefault();
			for(Entry<Pair<String,String>,Map<String,Long>> entry: map.entrySet()){
				String characterizationId = entry.getKey().getFirst();
				String characterizationDescription = entry.getKey().getSecond();
				XAttributeLiteral resourceLit = factory.createAttributeLiteral(characterizationId, characterizationDescription, this);
				
				for(Entry<String,Long> skill: entry.getValue().entrySet()){
					String skillId = skill.getKey();
					Long skillValue = skill.getValue();
					XAttributeDiscrete skillDis = factory.createAttributeDiscrete(skillId, skillValue, this);
					resourceLit.getAttributes().put(skillId, skillDis);
				}				
				attr.getAttributes().put(characterizationId, resourceLit);	
			}
			log.getAttributes().put(KEY_EXPERTISE_REQUIRED, attr);
		}
	}
	
	
	
	
	

	

}
